from hackle.evaluation.evaluator import Evaluator
from hackle.evaluation.flow.evaluation_flow_factory import EvaluationFlowFactory
from hackle import exceptions as hackle_exceptions
from hackle import logger as _logging
from hackle.decision import ExperimentDecision, DecisionReason, FeatureFlagDecision
from hackle.event.event_dispatcher import EventDispatcher
from hackle.event.event_processor import BatchEventProcessor
from hackle.internal_client import InternalClient
from hackle.workspace_fetcher import WorkspaceFetcher


def __singleton(cls):
    instance = [None]

    def wrapper(*args, **kwargs):
        if instance[0] is None:
            instance[0] = cls(*args, **kwargs)
        return instance[0]

    return wrapper


@__singleton
class Client(object):
    def __init__(self, sdk_key=None, logger=None, timeout=None):
        if sdk_key is None:
            raise hackle_exceptions.RequiredParameterException('sdk_key must not be empty.')

        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())

        self.internal_client = InternalClient(
            evaluator=Evaluator(EvaluationFlowFactory(self.logger)),
            workspace_fetcher=WorkspaceFetcher(sdk_key, logger=self.logger, timeout=timeout),
            event_processor=BatchEventProcessor(sdk_key, EventDispatcher(), self.logger),
            logger=self.logger
        )

    def close(self):
        self.internal_client.close()

    def __exit__(self):
        self.close()

    def variation(self, experiment_key, user, default_variation='A'):
        """
        Decide the variation to expose to the user for experiment.

        This method return the "A" if:
            - The experiment key is invalid
            - The experiment has not started yet
            - The user is not allocated to the experiment
            - The decided variation has been dropped

        :param int experiment_key: the unique key of the experiment.
        :param hackle.model.User user: the user to participate in the experiment.
        :param str default_variation: the default variation of the experiment.

        :return: the decided variation for the user, or the default variation.
        """
        return self.variation_detail(experiment_key, user, default_variation).variation

    def variation_detail(self, experiment_key, user, default_variation='A'):
        """
        Decide the variation to expose to the user for experiment, and returns an object that
        describes the way the variation was decided.

        :param int experiment_key: the unique key of the experiment.
        :param hackle.model.User user: the user to participate in the experiment.
        :param str default_variation: the default variation of the experiment.

        :return: a object describing the result
        """
        try:
            return self.internal_client.decide_experiment(experiment_key, user, default_variation)
        except Exception as e:
            self.logger.error(
                "Unexpected error while deciding variation of experiment[{}]: {}".format(experiment_key, str(e)))
            return ExperimentDecision(default_variation, DecisionReason.EXCEPTION)

    def is_feature_on(self, feature_key, user):
        """
        Decide whether the feature is turned on to the user.

        :param int feature_key: the unique key of the feature.
        :param hackle.model.User user: the user requesting the feature.
        :return: True if the feature is on
                  False if the feature is off
        """
        return self.feature_flag_detail(feature_key, user).is_on

    def feature_flag_detail(self, feature_key, user):
        """
        Decide whether the feature is turned on to the user, and returns an object that
        describes the way the value was decided.

        :param int feature_key: the unique key of the feature.
        :param hackle.model.User user: the user requesting the feature.

        :return: a object describing the result
        """
        try:
            return self.internal_client.decide_feature_flag(feature_key, user)
        except Exception as e:
            self.logger.error("Unexpected error while deciding feature flag[{}]: {}".format(feature_key, str(e)))
            return FeatureFlagDecision(False, DecisionReason.EXCEPTION)

    def track(self, event, user):
        """
        Records the event performed by the user with additional numeric value.

        :param hackle.model.Event event: the unique key of the event. MUST NOT be null.
        :param user: the identifier of user that performed the event. MUST NOT be null.
        """
        try:
            self.internal_client.track_event(event, user)
        except Exception as e:
            self.logger.error('Unexpected error while tracking event: {}'.format(str(e)))



class OverrideResolver(object):
    def __init__(self, target_matcher, action_resolver):
        self.target_matcher = target_matcher
        self.action_resolver = action_resolver

    def resolve_or_none(self, workspace, experiment, user):
        overridden_variation_id = experiment.user_overrides.get(user.id)
        if overridden_variation_id:
            return experiment.get_variation_by_id_or_none(overridden_variation_id)

        for overridden_rule in experiment.segment_overrides:
            if self.target_matcher.matches(overridden_rule.target, workspace, user):
                return self.action_resolver.resolve_or_none(overridden_rule.action, workspace, experiment, user)

        return None



from hackle.commons import validator, utils
from hackle.decision import ExperimentDecision, FeatureFlagDecision, DecisionReason
from hackle.entities import EventType
from hackle.event.user_event import ExposureEvent, TrackEvent


class InternalClient(object):

    def __init__(self, evaluator, workspace_fetcher, event_processor, logger):
        self.evaluator = evaluator
        self.workspace_fetcher = workspace_fetcher
        self.event_processor = event_processor
        self.logger = logger

    def close(self):
        self.workspace_fetcher.stop()
        self.event_processor.stop()

    def decide_experiment(self, experiment_key, user, default_variation_key):
        if not validator.is_non_zero_and_empty_int(experiment_key):
            self.logger.error('Experiment Key must not be empty. : {}'.format(experiment_key))
            return ExperimentDecision(default_variation_key, DecisionReason.INVALID_INPUT)

        if not validator.is_valid_user(user):
            self.logger.warning(
                'The user is not valid. user\'s type must be hackle.model.User and user.id\'s type must be '
                'string_types : {}'.format(user))
            return ExperimentDecision(default_variation_key, DecisionReason.INVALID_INPUT)

        if not validator.is_valid_properties(user.properties):
            self.logger.warning(
                'User properties is not valid. User properties must be not dic types and items types must be string_types, number, bool.'
                ': {}'.format(user.properties))
            user.properties = utils.filter_properties(user.properties)

        workspace = self.workspace_fetcher.get_workspace()
        if not workspace:
            self.logger.warning('Invalid Workspace. Hackle instance is not valid. {}'.format('variation'))
            return ExperimentDecision(default_variation_key, DecisionReason.SDK_NOK_READY)

        experiment = workspace.get_experiment_or_none(experiment_key)
        if not experiment:
            self.logger.debug('Could not find valid Experiment. "%s"' % experiment_key)
            return ExperimentDecision(default_variation_key, DecisionReason.EXPERIMENT_NOT_FOUND)

        evaluation = self.evaluator.evaluate(workspace, experiment, user, default_variation_key)
        self.event_processor.process(ExposureEvent(user, experiment, evaluation))

        return ExperimentDecision(evaluation.variation_key, evaluation.reason)

    def decide_feature_flag(self, feature_key, user):
        if not validator.is_non_zero_and_empty_int(feature_key):
            self.logger.error('Feature Key must not be empty. : {}'.format(feature_key))
            return FeatureFlagDecision(False, DecisionReason.INVALID_INPUT)

        if not validator.is_valid_user(user):
            self.logger.warning(
                'The user is not valid. user\'s type must be hackle.model.User and user.id\'s type must be '
                'string_types : {}'.format(user))
            return FeatureFlagDecision(False, DecisionReason.INVALID_INPUT)

        if not validator.is_valid_properties(user.properties):
            self.logger.warning(
                'User properties is not valid. User properties must be not dic types and items types must be string_types, number, bool.'
                ': {}'.format(user.properties))
            user.properties = utils.filter_properties(user.properties)

        workspace = self.workspace_fetcher.get_workspace()
        if not workspace:
            return FeatureFlagDecision(False, DecisionReason.SDK_NOK_READY)

        feature_flag = workspace.get_feature_flag_or_none(feature_key)
        if not feature_flag:
            return FeatureFlagDecision(False, DecisionReason.FEATURE_FLAG_NOT_FOUND)

        evaluation = self.evaluator.evaluate(workspace, feature_flag, user, 'A')
        self.event_processor.process(ExposureEvent(user, feature_flag, evaluation))

        if evaluation.variation_key == 'A':
            return FeatureFlagDecision(False, evaluation.reason)
        else:
            return FeatureFlagDecision(True, evaluation.reason)

    def track_event(self, event, user):
        if not validator.is_valid_event(event):
            self.logger.error('Event is not valid. Event must be hackle.model.event and event.id\'s type must be '
                              'string_types. value\'s type must be numeric. '
                              ': {}'.format(event))
            return

        if not validator.is_valid_user(user):
            self.logger.warning('User is not valid. user\'s type must be hackle.model.User and user.id\'s type must be '
                                'string_types : {}'.format(user))
            return

        if not validator.is_valid_properties(event.properties):
            self.logger.warning(
                'Event properties is not valid. Event properties must be not dic types and items types must be string_types, number, bool.'
                ': {}'.format(event.properties))
            event.properties = utils.filter_properties(event.properties)

        if not validator.is_valid_properties(user.properties):
            self.logger.warning(
                'User properties is not valid. User properties must be not dic types and items types must be string_types, number, bool.'
                ': {}'.format(user.properties))
            user.properties = utils.filter_properties(user.properties)

        event_type = self._event_type(event)
        self.event_processor.process(TrackEvent(user, event_type, event))

        return

    def _event_type(self, event):
        config = self.workspace_fetcher.get_workspace()

        if not config:
            return EventType(0, event.key)

        event_type = config.get_event_type_or_none(event.key)

        if not event_type:
            return EventType(0, event.key)

        return event_type
